// ---------------------------------------------------------------------
// Time-stamp: <Samstag, 25.01.2003, 16:58:12; edited by heiko>
// 
// Tina's random number generators TRNG
//
// Copyright (C) 2001, 2002 Heiko Bauke
//
// heiko.bauke@physik.uni-magdeburg.de
//
// TRNG is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation. This program
// is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// ---------------------------------------------------------------------

#if !defined TRNG_H

#define TRNG_H

#include <iostream>
#include <cmath>
#include <climits>
#include <vector>
#include <trnglib.h>

#define TRNG_VERSION "2"
#define TRNG_REVISION "2"


//! Tina's random number generators namespace
/*!
  All function and classes are encapsulated by the namespace TRNG.
*/
namespace TRNG {
  
  //! pseudo random number generator types
  /*!
    Every pseudo random number generator's type can be identified by a class
    member type which has anny value from this enumeration type.
  */
  enum RNG_type {
    RNG_t,          //!< not specialized random number generator 
    generic_MLCG_t, //!< generic multiplicative linear congruential random number generator
    ParkMiller_t,   //!< random number generator class ParkMiller
    LCG32_t,        //!< random number generator class LCG32
    LCG64_t,        //!< random number generator class LCG64
    MRG2_t,         //!< random number generator class MRG2
    MRG3_t,         //!< random number generator class MRG3
    MRG4_t,         //!< random number generator class MRG4
    YARNLCG64_t,    //!< random number generator class YARNLCG64
    YARN2_t,        //!< random number generator class YARN2
    YARN3_t,        //!< random number generator class YARN3
    YARN4_t,        //!< random number generator class YARN4
    CLCG2_t,        //!< random number generator class CLCG2
    CLCG3_t,        //!< random number generator class CLCG3
    CLCG4_t,        //!< random number generator class CLCG4
    EINV_t,         //!< random number generator class EINV
    EINVLCG64_t,    //!< random number generator class EINVLCG64
    trng_gsl_t,     //!< random number generator class trng_gsl
    user1_t=1001,   //!< user defined random number generator class nr 1
    user2_t=1002,   //!< user defined random number generator class nr 2
    user3_t=1003,   //!< user defined random number generator class nr 3
    user4_t=1004,   //!< user defined random number generator class nr 4
    user5_t=1005    //!< user defined random number generator class nr 5
  };   
 
  //! two dimensional vector structure
  struct vector2d_struct {
    //! 1st element
    double x1;
    //! 2nd element
    double x2;
  };

  //! two dimensional vector
  /*! 
    This structure is for storing two dimensional vectors. The method
    TRNG::RNG::spherical2d(void) returns this structure.
   */
  typedef struct vector2d_struct vector2d;
  
  //! three dimensional vector structure
  struct vector3d_struct {
    //! 1st element
    double x1;
    //! 2nd element
    double x2;
    //! 3rd element
    double x3;
  };

  //! three dimensional vector
  /*! 
    This structure is for storing three dimensional vectors. The method
    TRNG::RNG::spherical3d(void) returns this structure.
  */
  typedef struct vector3d_struct vector3d;

  //! four dimensional vector structure
  struct vector4d_struct {
    //! 1st element
    double x1;
    //! 2nd element
    double x2;
    //! 3rd element
    double x3;
    //! 4th element
    double x4;
  };

  //! four dimensional vector
  /*! 
    This structure is for storing four dimensional vectors. The method
    TRNG::RNG::spherical4d(void) returns this structure.
   */
  typedef struct vector4d_struct vector4d;

  // -------------------------------------------------------------------

  //! pseudo random number generator template
  /*! 
    All the pseudo random number generators are derived form this base
    class. Routines for generation of uniform and nonuniform variartes 
    are implemented in this base class. Generator specific tasks like 
    sequence splitting or leapfrog method are implementetd by the derived 
    classes. The method rand(), the generator's core method, has of 
    course also to be reimplemented for every new generator. 
    \author Heiko Bauke
  */
  template<class RNG_type> class RNG {
  private:
    RNG_type & derived(void) {
      return static_cast<RNG_type &>(*this);
    }
    
  protected:
    //! maximum value that the random number gernerator's method rand() can return 
    long max_val;
    //! the half of the maximum value that the random number gernerator's method rand() can return 
    long max_val2;
  public:
    //! pseudo random number generator type
    /*!	
      This numerical value determins the random number generator type.
    */
    static const TRNG::RNG_type type=RNG_t;
    
    //! pseudo random number generator's name
    /*! 
      Returns a pointer to a zero terminated string containing the pseudo 
      random number generator's name.
      \return pointer to a zero termiated string
    */
    const char * name(void) {
      return derived().name(); 
    }
    
    //! reset the pseudo random number generator
    /*! 
      The pseudo random number generator's parameters are set to some 
      default values.
    */
    void reset(void)  { 
      derived().reset(); 
    }
    
    //! seed the pseudo random number generator
    /*! 
      The pseudo random number generator's state is set. Calling reset()
      and seed() with the seed value used by the initialisation will
      set the generator in the same state as after initialisation.
      \param s new seed
      \exception error if \f$s<0\f$
      \sa TRNG::error
    */
    void seed(long s=0l) { 
      derived().seed(s);
    }
    
    //! next pseudo random number
    /*! 
      This is the generator's core method. It calculates the next pseudo
      random number. It's an integer number \f$r\f$ with 
      \f$0\le r\le\mathrm{max}\f$. You can determine the upper bound 
      \f$\mathrm{max}\f$ by calling the method max().	
      \return next integer pseudo random number
    */
    long rand(void) { 
      return derived().rand(); 
    }

    //! maximal pseudo random number
    /*!
      \return maximal pseudo random number returned by rand()
    */
    long max(void) {
      return max_val;
    };
    
    //! pseudo random boolean
    /*!
      \return pseudo random boolean with probability \f$\frac{1}{2}\f$ for 
      return value true
    */
    bool boolean(void) {
      return (rand()<=max_val2) ? true : false;
    }

    //! pseudo random boolean
    /*!
      \return pseudo random boolean with probability \f$p\f$ for return value 
      true
    */
    bool boolean(const double p) {
      return (rand()<static_cast<long>(p*max())+1.0) ? true : false;
    }
    
    //! pseudo random number
    /*!
      \return a in \f$[0,1)\f$ uniform distributed random number
    */
    double uniform(void) {
      return uniformco();
    }

    //! pseudo random number
    /*!
      \return a in \f$[a,b)\f$ uniform distributed random number
      \param a lower bound
      \param b upper bound
    */
    double uniform(const double a, const double b) {
      return uniformco(a, b);
    }

    //! pseudo random number
    /*!
      \return a in \f$[0,1)\f$ uniform distributed random number
    */
    double uniformco(void) {
      return static_cast<double>(rand())/
	(static_cast<double>(max())+1.0);
    }

    //! pseudo random number
    /*!
      \return a in \f$[a,b)\f$ uniform distributed random number
      \param a lower bound
      \param b upper bound
    */
    double uniformco(const double a, const double b) {
      return a+(b-a)*static_cast<double>(rand())/
	(static_cast<double>(max())+1.0);
    }

    //! pseudo random number
    /*!
      \return a in \f$[0,1]\f$ uniform distributed random number
    */
    double uniformcc(void) {
      return static_cast<double>(rand())/
	static_cast<double>(max());
    }
    
    //! pseudo random number
    /*!
      \return a in \f$[a,b]\f$ uniform distributed random number
      \param a lower bound
      \param b upper bound
    */
    double uniformcc(const double a, const double b) {
      return a+(b-a)*static_cast<double>(rand())/
	static_cast<double>(max());
    }

    //! pseudo random number
    /*!
      \return a in \f$(0,1]\f$ uniform distributed random number
    */
    double uniformoc(void) {
      return (static_cast<double>(rand())+1.0)/
	(static_cast<double>(max())+1.0);
    }
    
    //! pseudo random number
    /*!
      \return a in \f$(a,b]\f$ uniform distributed random number
      \param a lower bound
      \param b upper bound
    */
    double uniformoc(const double a, const double b) {
      return a+(b-a)*(static_cast<double>(rand())+1.0)/
	(static_cast<double>(max())+1.0);
    }
    
    //! pseudo random number
    /*!
      \return a in \f$(0,1)\f$ uniform distributed random number
    */
    double uniformoo(void) {
      return (static_cast<double>(rand())+1.0)/
	(static_cast<double>(max())+2.0);
    }

    //! pseudo random number
    /*!
      \return a in \f$(a,b)\f$ uniform distributed random number
      \param a lower bound
      \param b upper bound
    */
    double uniformoo(const double a, const double b) {
      return a+(b-a)*(static_cast<double>(rand())+1.0)/
	(static_cast<double>(max())+2.0);
    }
    
    //! pseudo random number
    /*!
      \return If \f$b>0\f$ a in \f$[0, b)\f$ uniform distributed natural random number is returned, if \f$b<0\f$ return value is in \f$(b, 0]\f$.
      \param b upper bound
    */
    long uniforml(const long b) {
      return static_cast<long>(uniform(0.0, static_cast<double>(b)));
    }
    
    //! pseudo random number
    /*!
      \return a in \f$[a, b)\f$ uniform distributed natural random number
      \param a lower bound
      \param b upper bound
    */
    long uniforml(const long a, const long b) {
      return static_cast<long>(uniform(static_cast<double>(a),
				       static_cast<double>(b)));
    }
    
    //! pseudo random number
    /*!
      A normal distributed random variate has a probability density
      \f[
      p(x)=\dfrac{1}{\sqrt {2\pi}\sigma}\mathrm{e}^{-\frac{x^2}{2\sigma^2}}\,.
      \f]
      This method uses the polar (Box-Mueller) method\latexonly, see 
      \cite{Knuth1981} for details\endlatexonly.
      \return a normal distributed random number with mean \f$\mu\f$ and variance
      \f$\sigma\f$ 
      \param sigma variance \f$\sigma\f$
      \param mu mean \f$\mu\f$
      \exception error if \f$\sigma\le 0\f$
      \sa TRNG::error
    */
    double normal_dist(const double sigma=1.0, const double mu=0.0) {
      // static double s, t1, t2;
      double s, t1, t2;
      // static bool calc=true;
      if (sigma<=0.0)
	throw error("negative or zero standard deviation in TRNG::RNG::normal_dist");
      // if (calc) {
      // calc=false;
      do {
	t1=uniformoo(-1.0, 1.0);
	t2=uniformoo(-1.0, 1.0);
	s=t1*t1+t2*t2;
      } while (s>=1.0);
      s=std::sqrt(-2.0*std::log(s)/s);
      return t1*s*sigma+mu;
      // } else {
      // calc=true;
      // return t2*s*sigma+mu;
      // }
    }

    //! pseudo random number
    /*!
      A exponential distributed random variate has a probability density
      \f[
      p(x)=
      \begin{cases}
      \frac{1}{\mu}\mathrm{e}^{-\frac{x}{\mu}} & x\ge 0\,,\qquad \mu>0 \\
      0 & \text{else}
      \end{cases}\,.
      \f]
      \latexonly
      The transformation method is discribed in \cite{Knuth1981}. 
      \endlatexonly
      One random number is used to get one random number with exponential 
      distribution.
      \return a exponential distributed random number with mean \f$\mu\f$
      \param mu mean \f$\mu\f$
      \exception error if \f$\mu\le 0\f$
      \sa TRNG::error
    */
    double exp_dist(const double mu=1.0) {
      if (mu<=0.0)
	throw error("negative or zero parameter in TRNG::RNG::exp_dist");
      double t=uniformoc();
      return -mu*std::log(t);
    }

    //! pseudorandom number
    /*!
      The two-sided exponential probability distribution is
      \f[
      p(x)=\dfrac{1}{2a}\mathrm{e}^{-|x|/a}\,,\qquad a>0\,.
      \f]
      It is also known as the Laplace distribution. The is implementation 
      adopted from the GNU scientific library and uses a simple 
      transformation method. One random number is used to get one random
      number with Laplace distribution.
      \return a pseudo random number with probability density \f$p(x)\f$
      \param a parameter \f$a\f$, \f$a>0\f$
      \exception error if \f$a\le0\f$
      \sa TRNG::error
    */
    double laplace_dist(const double a=1.0) {
      if (a<=0.0)
	throw error("parameter less or equal zero in TRNG::RNG::laplace_dist");
      double u;
      do {
	u=uniformoo(-1.0, 1.0);
      } while (u==0.0);
      return u<0.0 ? a*std::log(-u) : -a*std::log(u);
    }

    //! pseudorandom number
    /*!
      This method generates a pseudo randeom number with a tent shaped
      probability distribution.
      \f[
      p(x)=
      \begin{cases}
      \dfrac{x+a}{a^2} & -a\le x\le 0 \\[1.5ex]
      \dfrac{a-x}{a^2} & 0\le x\le a \\
      0 & \text{else}
      \end{cases}\,,
      \qquad\text{with }a>0
      \f]
      The implementation uses a simple transformation method. One random 
      number is used to get one random number with a tent shaped
      distribution.
      \return a pseudo random number with probability density \f$p(x)\f$
      \param a parameter \f$a\f$, \f$a>0\f$
      \exception error if \f$a\le0\f$
      \sa TRNG::error
    */
    double tent_dist(const double a=1.0) {
      if (a<=0.0)
	throw error("parameter less or equal zero in TRNG::RNG::tent_dist");
      double y=uniformoo();
      if (y<0.5)
	return (-1.0+std::sqrt(2*y))*a;
      else
	return (1.0-std::sqrt(2.0-2.0*y))*a;
    }
    
    //! pseudo random number
    /*!
      A gamma distributed random number has a probability density
      \f[
      p(x)=
      \begin{cases}
      \dfrac{1}{\Gamma(a)b^a} x^{a-1} \mathrm{e}^{-\frac{x}{b}} & x>0 \\
      0 & \text{else}
      \end{cases}\,.
      \f]
      The is implemetation adopted from the GNU scientific library.
      \latexonly
      See also \cite{Knuth1981}. 
      \endlatexonly
      \return a gamma distributed random number 
      \param a parameter \f$a\f$
      \param b parameter \f$b\f$
      \exception error if \f$a\le 0\f$ or \f$b\le 0\f$
      \sa TRNG::error
    */
    double Gamma_dist(const double a, const double b) {
      if (a<=0.0 || b<=0.0)
	throw error("parameter less or equal zero in TRNG::RNG::Gamma_dist");
      double a_int=std::floor(a);
      if (a_int==a) {
	// if a is an integer
	if (a<12) {
	  double prod=1.0;
	  for (int i=0; i<a; ++i)
	    prod*=uniformoo();
	  // Note: for 12 iterations we are safe against underflow, since
	  // the smallest positive random number is O(2^-32). This means
	  // the smallest possible product is 2^(-12*32) = 10^-116 which
	  // is within the range of double precision.
	  return -b*std::log(prod);
	} else {
	  // Works only if a>1, and is most efficient if a is large
	  // This algorithm, reported in Knuth, is attributed to Ahrens.  A
	  // faster one, we are told, can be found in: J. H. Ahrens and
	  // U. Dieter, Computing 12 (1974) 223-246.
	  double sqa, x, y, v;
	  sqa=std::sqrt(2.0*a-1.0);
	  do {
	    do {
	      const double pi=3.14159265358979324;
	      y=std::tan(pi*uniformco());
	      x=sqa*y+a-1.0;
	    }
	    while (x<=0.0);
	    v=uniformco();
	  }
	  while (v>(1.0+y*y)*std::exp((a-1.0)*log(x/(a-1))-sqa*y));
	  return b*x;
	}
      }
      if (a_int==0.0) {
	// if a<1
	// This is exercise 16 from Knuth; see page 135, and the solution is
	// on page 551.
	const double e=2.71828182844;
	double p, q, x, u, v;
	p=e/(a+e);
	do {
	  u=uniformco();
	  v=uniformoo();
	  if (u<p) {
	    x=std::exp((1.0/a)*std::log(v));
	    q=std::exp(-x);
	  } else {
	    x=1.0-std::log(v);
	    q=std::exp((a-1.0)*std::log(x));
	  }
	} while (uniformco()>=q);
	return b*x;
      }
      return Gamma_dist(a_int, b)+Gamma_dist(a-a_int, b); 
    }

    //! pseudo random number
    /*!
      A Beta distributed random number has a probability density
      \f[
      p(x)=
      \begin{cases}
      \dfrac{\Gamma(a+b)}{\Gamma(a)\Gamma(b)}x^{a-1}(1-x)^{b-1} &
      0\le x\le 1 \\
      0 & \text{else}
      \end{cases}
      \f]
      \latexonly
      The method is discribed in \cite{Knuth1981} pp.~129-130.
      \endlatexonly
      \return a Beta distributed random number 
      \param a parameter \f$a\f$, \f$a>0\f$
      \param b parameter \f$b\f$, \f$b>0\f$
      \exception error if \f$a\le 0\f$ or \f$b\le 0\f$
      \sa TRNG::error
    */
    double Beta_dist(const double a, const double b) {
      if (a<=0.0 || b<=0.0)
	throw error("negative or zero parameter in TRNG::RNG::Beta_dist");
      double t1=Gamma_dist(a, 1.0);
      double t2=Gamma_dist(b, 1.0);
      return t1/(t1+t2);
    }

    //! pseudo random number
    /*!
      The \f$\chi^2\f$-distribution is just a special case of the 
      \f$\Gamma\f$-distribution with \f$a=\nu/2\f$ and \f$b=1\f$.
      \f[
      p(x)=\frac{\left(\frac{x}{2}\right)^{\frac{\nu}{2}-1}}
      {2\Gamma(\frac{\nu}{2})}
      \mathrm{e}^{-\frac{x}{2}}\,,\qquad x\le0
      \f]
      \latexonly
      The method is discribed in \cite{Knuth1981} p.~130.
      \endlatexonly
      \return a chi square distributed random number
      \param nu degrees of freedom \f$\nu\f$
      \exception error if \f$\nu<1\f$
      \sa TRNG::error
    */
    double chi_square_dist(const double nu) {
      if (nu<1.0)
	throw error("parameter less than one in TRNG::RNG::chi_square_dist");
      return 2.0*Gamma_dist(0.5*nu, 1.0);
    }

    //! pseudo random number
    /*!
      The binomial distribution is a discrete distribution with
      \f[
      p(k)=\frac{n!}{k!(n-k)!}p^k(1-p)^{n-k}\,,\qquad 0\le k\le n\,.
      \f]
      \latexonly
      The method is discribed in \cite{Knuth1981} p.~131. 
      \endlatexonly 
      The implemetation is adopted from the GNU scientific library.
      \return a binomial distributed pseudo random number
      \param n number of trails \f$n\f$
      \param p probability \f$p\f$ in each trail
      \exception error if \f$p\le 0\f$ or \f$p>1\f$ or \f$n\le 0\f$
      \sa TRNG::error
    */
    long binomial_dist(long n, double p=0.5) {
      if (p<=0.0 || p>1.0)
	throw error("probability <=0.0 or >1.0 in TRNG::RNG::binomial_dist");
      if (n<1l)
	throw error("less than one trail TRNG::RNG::binomial_dist");
      long i, a, b, k=0;
      while (n>12l) {  // This parameter is tunable
	double X;
	a=1l+(n/2l);
	b=1l+n-a;
	X=Beta_dist(static_cast<double>(a), static_cast<double>(b));
	if (X>=p) {
	  n=a-1l;
	  p/=X;
	} else {
	  k+=a;
	  n=b-1l;
	  p=(p-X)/(1l-X);
	}
      }
      for (i=0l; i<n; i++)
	if (uniformco()<p)
	  k++;
      return k;
    }

    //! pseudo random number
    /*!
      The binomial distribution is a discrete distribution with
      \f[
      p(k)=\frac{n!}{k!(n-k)!}p^k(1-p)^{n-k}\,,\qquad 0\le k\le n\,.
      \f]
      This method's implementation uses a lookup table and is very fast
      if this method is called often with the same parameter set.
      \return a binomial distributed pseudo random number
      \param n number of trails \f$n\f$
      \param p probability \f$p\f$ in each trail
      \exception error if \f$p\le 0\f$ or \f$p>1\f$ or \f$n\le 0\f$
      \sa TRNG::error
    */
    long binomial_dist_tab(long n, double p=0.5) {
      if (p<=0.0 || p>1.0)
	throw error("probability <=0.0 or >1.0 in TRNG::RNG::binomial_dist_tab");
      if (n<1l)
	throw error("less than one trail TRNG::RNG::binomial_dist_tab");
      static long n_=0l;
      static double p_=0.0;
      static std::vector<double> prob;
      if (n!=n_ || p!=p_) {
	n_=n;
	p_=p;
	prob.resize(n_);
	// use
	// ( n )            (  n  )   n-k ( n )
	// (   ) = 1  and   (     ) = --- (   )
	// ( 0 )            ( k+1 )   k+1 ( k )
	double b=1.0;
	for (int k=0l; k<n_; ++k) {
	  prob[k]=b*std::pow(p_, static_cast<double>(k))*
	    std::pow(1.0-p_, static_cast<double>(n_-k));
	  if (k>0l)
	    prob[k]+=prob[k-1l];
	  b=std::floor(b*static_cast<double>(n_-k)/
		       static_cast<double>(k+1l)+0.5); 
	}
      }
      return find_interval(prob, uniformco());
    }

    //! pseudo random number
    /*!
      Student's \f$t\f$-distribution with \f$\nu\f$ degrees of freedom is 
      defined as 
      \f[
      p(x)=\frac{\Gamma\left(\frac{\nu-1}{2}\right)}
      {\sqrt{\nu\pi}\Gamma\left(\frac{\nu}{2}\right)}
      \left(1+\frac{x^2}{\nu}\right)^{-\frac{\nu+1}{2}}\,.
      \f]
      This method's implementation is adopted from the GNU scientific 
      library\latexonly, see also \cite{Knuth1981}\endlatexonly.
      \return a random number with Student's-\f$t\f$ distribution with 
      \f$\nu\f$ degrees of freedom
      \param nu degrees of freedom, \f$\nu\,>\,0\f$
      \exception error if \f$\nu\le 0\f$
      \sa TRNG::error
    */
    double Student_t_dist(const double nu) {
      if (nu<=0.0)
	throw error("parameter less than or equal zero in TRNG::RNG::Student_t_dist");
      if (nu<=2.0) 
	return normal_dist()/std::sqrt(chi_square_dist(nu)/nu);
      else {
	double y1, y2, z;
	do {
	  y1=normal_dist();
	  y2=exp_dist(1.0/(0.5*nu-1.0));
	  z=y1*y1/(nu-2.0);
	} while (1.0-z<0.0 || std::exp(-y2-z)>(1.0-z));
	// Note that there is a typo in Knuth's formula, the line below
	// is taken from the original paper of Marsaglia, Mathematics of
	// Computation, 34 (1980), p 234-256 
	return y1/std::sqrt((1.0-2.0/nu)*(1.0-z));
      }
    }

    //! pseudo random number
    /*!
      The probability distribution for Poisson variates is
      \f[
      p(k)=\frac{\mu^k}{k!}\mathrm{e}^{-\mu}\,,\qquad k\ge 0\,.
      \f]
      This method's implementation is adopted from the GNU scientific 
      library\latexonly, see also \cite{Knuth1981}\endlatexonly.
      \return a poisson distributed pseudo random number
      \param mu mean \f$\mu\f$
      \exception error if \f$\mu \le 0\f$
      \sa TRNG::error
    */
    long poisson_dist(double mu=1.0) {
      if (mu<=0)
	throw error("parameter less than 0 in TRNG::RNG::poisson_dist");
      double emu;
      double prod=1.0;
      long k=0;  
      while (mu>10.0) {
	long m=static_cast<long>(mu*(7.0/8.0));
	double X=Gamma_dist(static_cast<long>(m), 1.0);
	if (X>=mu) {
	  return k+binomial_dist(m-1l, mu/X);
	} else {
	  k+=m;
	  mu-=X; 
	}
      }
      // This following method works well when mu is small
      emu=std::exp(-mu);
      do {
	prod*=uniformco();
	k++;
      } while (prod>emu);
      return k-1l;
    }

    //! pseudo random number
    /*!
      The geometric probability distribution is
      \f[
      p(k)=q(1-q)^{k-1}\,,\qquad k\ge 1\,.
      \f]
      This method's implementation is adopted from the GNU scientific 
      library\latexonly, see also \cite{Knuth1981}\endlatexonly.
      \return a geometric distributed pseudo random number
      \param q probability \f$q\f$
      \exception error if \f$q\le 0\f$ or \f$q>1\f$
      \sa TRNG::error
    */
    long geometric_dist(double q) {
      if (q<=0.0 ||q>1.0)
	throw error("parameter out of range in TRNG::RNG::geometric_dist");
      if (q==1.0)
	return 1l;
      return static_cast<long>
	(std::log(uniformoo())/std::log(1.0-p))+1l;
    }

    //! pseudo random number
    /*!
      Returns a random number calculated by the rejection method. Assume
      your desired probability distribution is \f$p(x)=\frac{3}{4}(1-x^2)\f$
      for \f$x\in [-1, 1]\f$. Write a class that calculates this probability 
      distribution
\verbatim
class p {
public:
  double operator()(double x) {
    return 0.75*(1.0-x*x);
  }
};
\endverbatim
      and use
\verbatim
my_rng.rejection(p(), -1.0, 1.0, 0.75)
\endverbatim
      to generate a pseudo random number with probability distribution 
      \f$p(x)\f$.
      \return random number
      \param p function object, a function describing the probability density in the range between \f$a_1\f$ and \f$a_2\f$
      \param a1 lower bound \f$a_1\f$
      \param a2 upper bound \f$a_2\f$
      \param p_max maximum of the probability function \f$p(x)\f$ for 
      \f$x\in [a_1, a_2]\f$
    */
    template<class t_function>
      double rejection(t_function p, double a1, double a2, double p_max)  {
      double t1, t2;
      do {
	t1=uniformco(a1, a2);
	t2=uniformco(0.0, p_max);
      } while (p(t1)<t2);
      return t1;
    }

    //! pseudo random number
    /*!
      This method can be used to generate discrete random variates with
      an arbitrary probability distribution. The algorithm is \f$O(\ln n)\f$.
      \latexonly
      If you need a faster algorithm see \cite{Walker1977}.
      \endlatexonly
      \return random number \f$k\f$, \f$0\le k \le n-1\f$
      \param p \f$n\f$ dimensional commulative probability vector \f$p_0\f$, 
      \f$p_1,\ldots\f$ \f$p_{n-1}\f$, with \f$p_i\,<\,p_{i+1}\f$ and \f$p_{n-1}=1\f$
      \exception error if vector empty
      \sa TRNG::error
    */
    long discrete_dist(const std::vector<double> p) {
      if (p.empty())
	throw error("empty vector in TRNG::discrete_dist");
      double x=uniformco();
      if (x<p[0])
	return 0l;
      long i1=0l;
      long i2=p.size()-1l;
      while (i2-i1>1l) {
	long i3=(i2+i1)/2;
	if (x<=p[i3])
	  i2=i3;
	else
	  i1=i3;
      }
      return i2;
    }

    //! pseudo random vector
    /*!
      This method calculates a unit vector with a uniform distributed 
      direction in two dimensions. For a vector uniform distributed inside
      the unit citcle multiply the vector with \f$\sqrt{u}\f$ where \f$u\f$ is
      uniforum distributed in \f$[0,1)\f$. The vector is stored in a structure
      #vector2d.
      \return a two dimensional unit vector 
    */
    vector2d spherical2d(void) {
      double r2, t1, t2, t12, t22;
      vector2d vec;
      do {
	t1=2.0*uniformco()-1.0;
	t2=uniformco();
	t12=t1*t1;
	t22=t2*t2;
	r2=t12+t22;
      } while (r2>1.0);
      vec.x1=2.0*t1*t2/r2;
      vec.x2=(t12-t22)/r2;
      return vec;
    }

    //! pseudo random vector
    /*!
      This method calculates a unit vector with a uniform distributed 
      direction in three dimensions. For a vector uniform distributed 
      inside the unit shere multiply the vector with \f$\sqrt[3]{u}\f$ where 
      \f$u\f$ is uniforum distributed in \f$[0,1)\f$. The vector is stored in 
      a structure #vector3d.
      \return a three dimensional unit vector 
    */
    vector3d spherical3d(void)  {
      double q, r2, t1, t2;
      vector3d vec;
      do {
	t1=2.0*uniformco()-1.0;
	t2=2.0*uniformco()-1.0;
	r2=t1*t1+t2*t2;
      } while (r2>1.0);
      q=2.0*sqrt(1.0-r2);
      vec.x1=t1*q;
      vec.x2=t2*q;
      vec.x3=1-2*r2;
      return vec;
    }

    //! pseudo random vector
    /*!
      This method calculates a unit vector with a uniform distributed 
      direction in four dimensions. For a vector uniform distributed inside 
      the unit hyper-shere multiply the vector with \f$\sqrt[4]{u}\f$ where 
      \f$u\f$ is uniforum distributed in \f$[0,1)\f$. The vector is stored in 
      a structure #vector4d.
      \return a four dimensional unit vector 
    */
    vector4d spherical4d(void)  {
      double q, r21, r22, t1, t2, t3, t4;
      vector4d vec;
      do {
	t1=2.0*uniformco()-1.0;
	t2=2.0*uniformco()-1.0;
	r21=t1*t1+t2*t2;
      } while (r21>1.0);
      do {
	t3=2.0*uniformco()-1.0;
	t4=2.0*uniformco()-1.0;
	r22=t3*t3+t4*t4;
      } while (r22>1.0);
      q=sqrt((1.0-r21)/r22);
      vec.x1=t1;
      vec.x2=t2;
      vec.x3=t3*q;
      vec.x4=t4*q;
      return(vec);
    }

    //! sequence splitting
    /*!
      The pseudo random number generator's sequence is splitted into \f$s\f$ 
      sequences using the leapfrog method. Sequence number \f$n\f$ is selected.
      \f$0\le n <s\f$
      \param s number of sequences
      \param n selected sequence
      \exception error if \f$s<1\f$ or \f$n\ge s\f$ or \f$n<0\f$
      \sa TRNG::error
    */
    void split(long s, long n) {
      derived().split(s, n);
    }

    //! sequence splitting
    /*!
      The pseudo random number generator jumps \f$s\f$ steps ahead.
      \param s determines the jump size
      \exception error if \f$s<0\f$
      \sa TRNG::error
    */
    void jump(long long s) {
      if (s<0ll)
	throw error("invalid argument for TRNG::RNG::jump");
      long i=0l;
      while (s>0ll) {
	if (s%2==1)
	  jump2(i);
	++i;
	s>>=1;
      }
    }

    //! sequence splitting
    /*!
      The pseudo random number generator jumps \f$n\cdot s\f$ steps ahead.
      \param s determines the jump size
      \exception error if \f$s<0\f$ or \f$n<0\f$ 
      \param s paremeter \f$s\f$
      \param n paremeter \f$n\f$
      \sa TRNG::error
    */
    void jump(long long s, long n) {
      if (s<0ll || n<0l)
	throw error("invalid argument for TRNG::RNG::jump");
      while (n>0l) {
	jump(s);
	--n;
      }
    }

    //! sequence splitting
    /*!
      The pseudo random number generator jumps \f$2^s\f$ steps ahead.
      \param s determines the jump size
      \exception error if \f$s<0\f$
      \sa TRNG::error
    */
    void jump2(long s) {
      derived().jump2(s);
    }

    //! sequence splitting
    /*!
      The pseudo random number generator jumps \f$n\cdot 2^s\f$ steps ahead.
      \param s determines the jump size
      \param n determines the jump size
      \exception if \f$s<0\f$ or \f$n<0\f$
      \sa TRNG::error
    */
    void jump2(long s, long n) {
      if (s<0l || n<0l)
	throw error("invalid argument for TRNG::RNG::jump2");
      while (n>0l) {
	if (n%2l==1l)
	  jump2(s);
	s+=1l;
	n>>=1;
      }
    }

    //! status saving
    /*!
      The status of the pseudo random number generator is saved into a
      vector.
      \param s reference to a vector of long
    */
    void save_status(std::vector<long> &s) {
      derived().save_status(s);
    }

    //! status restoring
    /*!
      The status of the pseudo random number generator is restored from a
      vector.
      \param s reference to a vector of long
    */
    void load_status(const std::vector<long> &s) {
      derived().load_status(s);
    }

//     //! generic copy constructor
//     /*!
//       This is the random number generator copy constructor. Useful if
//       some generators with the same sequence of random numbers are needed.
//      */
//     RNG<RNG_type> & RNG<RNG_type>::operator=(RNG<RNG_type> &other) {
//       if (this!=&other) {
// 	std::vector<long> s;
// 	other.save_status(s);
// 	load_status(s);
//       }
//       return *this;
//     }
    
  };

  // --------------------------------------------------------------------

  //! linear congruential generator
  /*!
    This class implements a simple linear congruential pseudo random number 
    generator with a prime modulus in the form
    \f[
    \begin{split}
    r_{i} & =a\cdot r_{i-1}\mod m \,.
    \end{split}
    \f]
    \f$r_i\f$ is the actual pseudo random number. The modulus \f$m\f$ has 
    to be a prime smaller than \f$2^{31}\f$ and \f$a\f$ a generating element 
    of the multiplicative group modulo \f$m\f$ to generate a maxmal length 
    period. The behaviour of method rand() is different from other generators,
    it returns never 0. For this reason this generator is for internal use 
    only. 
    \author Heiko Bauke
  */
  class generic_MLCG : public RNG<generic_MLCG> {
    // generic class for multiplicative linear congruential 
    // random number generators
  private:
    long a, a_save;
    long modulus, modulus_save;
    long r;
    void backward(void);
  public:
    static const RNG_type type=generic_MLCG_t;
    const char * name(void);    
    void reset(void);
    void seed(long s=0l);
    long rand(void);
    void split(long s, long n);    
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    generic_MLCG & generic_MLCG::operator=(RNG<generic_MLCG> &other);
    generic_MLCG(long a_=16807l, long modulus_=2147483647l, long seed_=0l);
  };

  inline long generic_MLCG::rand(void) {
    r=static_cast<long>
      ((static_cast<long long>(a)*static_cast<long long>(r))%modulus);
    return r;
  }

  // -------------------------------------------------------------------

  //! linear congruential generator
  /*!
    This class implements a simple linear congruential pseudo random number 
    generator with a prime modulus in the form
    \f[
    \begin{split}
    q_{i} & =a\cdot q_{i-1}\mod m \\
    r_{i} & = q_{i}-1\,.
    \end{split}
    \f]
    \f$r_i\f$ is the actual pseudo random number. The modulus \f$m\f$ has 
    to be a prime smaller than \f$2^{31}\f$ and \f$a\f$ a generating element 
    of the multiplicative group modulo \f$m\f$ to generate a maxmal length 
    period.
    \author Heiko Bauke
  */
  class ParkMiller : public RNG<ParkMiller> {
  private:
    generic_MLCG R;
  public:
    static const RNG_type type=ParkMiller_t;
    const char * name(void);    
    void reset(void);
    void seed(long s=0l);
    long rand(void);
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    ParkMiller & ParkMiller::operator=(RNG<ParkMiller> &other); 
    //! constructor
    /*! 
      The constructor's default values implement a modified random number 
      generator proposed by Park and Miller. This generator has a period of
      \f$2^{31}-2\approx 2^{31}\approx 2.15\cdot 10^9\f$. 
      \latexonly See also \cite{ParkMiller1988}. \endlatexonly
      \param a_ multiplier \f$a\f$ 
      \param modulus_ prime modulus \f$m\f$
      \param seed_ default seed
    */
    ParkMiller(long a_=16807l, long modulus_=2147483647l, long seed_=0l);
  };

  inline long ParkMiller::rand(void) {
    return R.rand()-1l;
  }

  
  // -------------------------------------------------------------------

  //! linear congruential generator
  /*!
    This class implements a simple linear congruential pseudo random number 
    generator with a power of two modulus in the form
    \f[
    \begin{split}
    q_{i} & = a\cdot q_{i-1}+b\mod 2^{32} \\
    r_{i} & = \lfloor q_{i}/2 \rfloor \,.
    \end{split}
    \f]
    \f$r_i\f$ is the actual pseudo random number. To get a full period of 
    \f$2^{32}\f$ \f$a\f$ and \f$b\f$ have to be choosen that 
    \f$a\equiv 1\mod 4\f$ and \f$b\f$ is odd.
    \author Heiko Bauke
  */
  class LCG32 : public RNG<LCG32> {
  private:
    unsigned long a, a_save;
    unsigned long b, b_save;
    unsigned long r;
    void backward(void);
  public:
    static const RNG_type type=LCG32_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);    
    long rand(void);    
    void split(long s, long n);    
    void jump2(long s);    
    void save_status(std::vector<long> &s);    
    void load_status(const std::vector<long> &s);
    LCG32 & LCG32::operator=(RNG<LCG32> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator used on VAX. This generator has a period of 
      \f$2^{32}\approx 4.29\cdot 10^9\f$. This generator is just a toy.
      Its period is too short and it has some bad statistical properties.
      \param a_ multiplier \f$a\f$ 
      \param b_ additive constant
      \param seed_ default seed
    */
    LCG32(unsigned long a_=69069ul, unsigned long b_=1ul, long seed_=0l);
  };

  inline long LCG32::rand(void) {
    r=a*r+b;
#if ULONG_MAX == 0xfffffffful
#else
    r&=0xfffffffful;
#endif
    return static_cast<long>(r>>1);
  }

  // -------------------------------------------------------------------

  //! linear congruential generator
  /*!
    This class implements a simple linear congruential pseudo random number 
    generator with a power of two modulus in the form
    \f[
    \begin{split}
    q_{i} & = a\cdot q_{i-1}+b\mod 2^{64} \\
    r_{i} & = \lfloor q_{i}/2^{33} \rfloor \,.
    \end{split}
    \f]
    \f$r_i\f$ is the actual pseudo random number. To get a full period of 
    \f$2^{64}\f$ \f$a\f$ and \f$b\f$ have to be choosen that \f$a\equiv 1\mod 4\f$ and 
    \f$b\f$ is odd.
    \author Heiko Bauke
   */
  class LCG64 : public RNG<LCG64> {
  private:
    unsigned long long a, a_save;
    unsigned long long b, b_save;
    unsigned long long r;
    void backward(void);
  public:
    static const RNG_type type=LCG64_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);
    long rand(void);
    void split(long s, long n);
    void jump2(long s);    
    void save_status(std::vector<long> &s);    
    void load_status(const std::vector<long> &s);
    LCG64 & LCG64::operator=(RNG<LCG64> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a=18\,145\,460\,002\,477\,866\,997\f$ and \f$b=1\f$.
      This generator has a period of \f$2^{64}\approx 1.84\cdot 10^{19}\f$.
      LCG64 is the quick and dirty generator in TRNG.
      \param a_ multiplier \f$a\f$ 
      \param b_ additive constant
      \param seed_ default seed
    */
    LCG64(unsigned long long a_=18145460002477866997ull, 
	  unsigned long long b_=1ul, 
	  long seed_=0ul);    
  };
  
  inline long LCG64::rand(void) {
    r=a*r+b;
    return static_cast<long>(r>>33);
  }

  // -------------------------------------------------------------------

  //! multiple recursive generator
  /*!
    This multiple recursive generator uses a linear recurrence of order 
    two with a prime modulus.
    \f[
    r_{i} = a_1\cdot r_{i-1}+a_2\cdot r_{i-2}\mod m
    \f]
    \author Heiko Bauke
  */
  class MRG2 : public RNG<MRG2> {
  private:
    long a0, a1, a0_save, a1_save;
    long r0, r1;
    long modulus, modulus_save;
    void backward(void);
  public:
    static const RNG_type type=MRG2_t;
    const char * name(void);    
    void reset(void);
    void seed(long s=0l);
    long rand(void);    
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);    
    MRG2 & MRG2::operator=(RNG<MRG2> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a_1=1\,498\,809\,829\f$, \f$a_2=1\,160\,990\,996\f$ 
      and \f$m=2^{31}-1\f$ 
      \latexonly as proposed in \cite{lEcuyer1993}\endlatexonly. 
      This generator has a period of 
      \f$2^{2\cdot 31}-1\approx 2^{62}\approx 4.61\cdot 10^{18}\f$.
      \param a0_ multiplier \f$a_1\f$ 
      \param a1_ multiplier \f$a_2\f$ 
      \param modulus_ prime modulus
      \param seed_ default seed
    */
    MRG2(long a0_=1498809829l, long a1_=1160990996l, 
	 long modulus_=2147483647l, long seed_=0l);
  };

  inline long MRG2::rand(void) {
    long t=static_cast<long>((static_cast<long long>(a0)*
			      static_cast<long long>(r0)+
			      static_cast<long long>(a1)*
			      static_cast<long long>(r1))%modulus);
    r1=r0;  r0=t;
    return r0;
  }

  // -------------------------------------------------------------------

  //! multiple recursive generator
  /*!
    This multiple recursive generator uses a linear recurrence of order 
    three with a prime modulus.
    \f[
    r_{i} = a_1\cdot r_{i-1}+a_2\cdot r_{i-2}+a_3\cdot r_{i-3}\mod m
    \f]
    \author Heiko Bauke
  */
  class MRG3 : public RNG<MRG3> {
  private:
    long a0, a1, a2, a0_save, a1_save, a2_save;
    long r0, r1, r2;
    long modulus, modulus_save;
    void backward(void);
  public:
    static const RNG_type type=MRG3_t;
    const char * name(void);
    void reset(void);    
    void seed(long s=0l);
    long rand(void);   
    void split(long s, long n);
    void jump2(long s);    
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    MRG3 & MRG3::operator=(RNG<MRG3> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a_1=2\,021\,422\,057\f$, \f$a_2=1\,826\,992\,351\f$,
      \f$a_3=1\,977\,753\,457\f$ and \f$m=2^{31}-1\f$ 
      \latexonly as proposed in \cite{lEcuyer1993}\endlatexonly. 
      This generator has a period of 
      \f$2^{3\cdot 31}-1\approx 2^{93}\approx 9.90\cdot 10^{27}\f$.
      \param a0_ multiplier \f$a_1\f$ 
      \param a1_ multiplier \f$a_2\f$ 
      \param a2_ multiplier \f$a_3\f$ 
      \param modulus_ prime modulus
      \param seed_ default seed
    */
    MRG3(long a0_=2021422057l, long a1_=1826992351l, 
	 long a2_=1977753457l, long modulus_=2147483647l, 
	 long seed_=0l);
  };

  inline long MRG3::rand(void) {
    long t=static_cast<long>((static_cast<unsigned long long>(a0)*
			      static_cast<unsigned long long>(r0)+
			      static_cast<unsigned long long>(a1)*
			      static_cast<unsigned long long>(r1)+
			      static_cast<unsigned long long>(a2)*
			      static_cast<unsigned long long>(r2))%modulus);
    r2=r1;  r1=r0;  r0=t;
    return r0;
  }
    
  // -------------------------------------------------------------------

  //! multiple recursive generator
  /*!
    This multiple recursive generator uses a linear recurrence of order 
    four with a prime modulus.
    \f[
    r_{i} = a_1\cdot r_{i-1}+a_2\cdot r_{i-2}+
    a_3\cdot r_{i-3}+a_4\cdot r_{i-4}\mod m
    \f]
    \author Heiko Bauke
  */
  class MRG4 : public RNG<MRG4> {
  private:
    long a0, a1, a2, a3, a0_save, a1_save, a2_save, a3_save;
    long r0, r1, r2, r3;
    long modulus, modulus_save;
    void backward(void);
  public:
    static const RNG_type type=MRG4_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);
    long rand(void);
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    MRG4 & MRG4::operator=(RNG<MRG4> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a_1=2\,001\,982\,722\f$, \f$a_2=1\,412\,284\,257\f$,
      \f$a_3=1\,155\,380\,217\f$, \f$a_4=1\,668\,339\,922\f$ and 
      \f$m=2^{31}-1\f$ 
      \latexonly as proposed by \cite{lEcuyer1993}\endlatexonly.
      This generator has a period of 
      \f$2^{4\cdot 31}-1\approx 2^{124}\approx 2.13\cdot 10^{37}\f$.
      \param a0_ multiplier \f$a_1\f$ 
      \param a1_ multiplier \f$a_2\f$ 
      \param a2_ multiplier \f$a_3\f$ 
      \param a3_ multiplier \f$a_4\f$ 
      \param modulus_ prime modulus
      \param seed_ default seed
    */
    MRG4(long a0_=2001982722l, long a1_=1412284257l, 
	 long a2_=1155380217l, long a3_=1668339922l,
	 long modulus_=2147483647l, long seed_=0l);
  };

  inline long MRG4::rand(void) {
    long t=static_cast<long>((static_cast<unsigned long long>(a0)*
			      static_cast<unsigned long long>(r0)+
			      static_cast<unsigned long long>(a1)*
			      static_cast<unsigned long long>(r1)+
			      static_cast<unsigned long long>(a2)*
			      static_cast<unsigned long long>(r2)+
			      static_cast<unsigned long long>(a3)*
			      static_cast<unsigned long long>(r3))%modulus);
    r3=r2;  r2=r1;  r1=r0;  r0=t;
    return r0;
  }

  // -------------------------------------------------------------------

  //! YARN -- a modified linear recursive generator
  /*!
    This class implements a simple linear congruential pseudo random number 
    generator with a power of two modulus in the form
    \f[
    \begin{split}
    q_{i} & = a\cdot q_{i-1}+b\mod 2^{64} \\
    r_{i} & = \lfloor q_{i}/2^{33} \rfloor \,.
    \end{split}
    \f]
    \f$r_i\f$ is the actual pseudo random number. To get a full period of 
    \f$2^{64}\f$ \f$a\f$ and \f$b\f$ have to be choosen that \f$a\equiv 1\mod 4\f$ and 
    \f$b\f$ is odd.
    Linear structures are destroyed by a nonlinear bijective mapping from
    \f$r_{i} = 
    \begin{cases}
    b^{q_{i}}\mod 2^{31}-1 & \text{if }q_{i}\ne 0 \wedge q_{i}\ne 2^{31}-1\\
    0                      & \text{else}  \\
    \end{cases}\,.
    \f$
    The parameter \f$b\f$ has to be a generating element of the 
    multiplicative group modulo \f$2^{31}-1\f$.
    \author Heiko Bauke
   */
  class YARNLCG64 : public RNG<YARNLCG64> {
  private:
    LCG64 R;
    power b;
    long b_save;
  public:
    static const RNG_type type=YARNLCG64_t;
    const char * name(void);    
    void reset(void);
    void seed(long s=0l);
    long rand(void);    
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    YARNLCG64 & YARNLCG64::operator=(RNG<YARNLCG64> &other); 
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a=18\,145\,460\,002\,477\,866\,997\f$ and \f$b=1\f$.
      This generator has a period of \f$2^{64}\approx 1.84\cdot 10^{19}\f$.
      \param a_ multiplier \f$a\f$ 
      \param b_ additive constant
      \param seed_ default seed
      \param c_ parameter \f$c\f$
    */
    YARNLCG64(unsigned long long a_=18145460002477866997ull, 
	      unsigned long long b_=1ul, 
	      long seed_=0ul, long c_=123567893l);
  };

  inline long YARNLCG64::rand(void) {
    long t=R.rand();
    return (t==0 || t==2147483647l) ? t : b(t);
  }

  // -------------------------------------------------------------------

  //! YARN -- a modified multiple recursive generator
  /*!
    This multiple recursive generator uses a linear recurrence of order 
    two with a prime modulus.     
    \f[
    q_{i} = a_1\cdot q_{i-1}+a_2\cdot q_{i-2}\mod m
    \f]
    Linear structures are destroyed by a nonlinear bijective mapping from
    \f$r_{i} = 
    \begin{cases}
    b^{q_{i}}\mod m & \text{if }q_{i}\ne 0 \\
    0         & \text{if }q_{i} =0 \\
    \end{cases}\,.
    \f$
    The parameter \f$b\f$ has to be a generating element of the 
    multiplicative group modulo \f$m\f$.
    \author Heiko Bauke
   */
  class YARN2 : public RNG<YARN2> {
  private:
    MRG2 R;
    power b;
    long b_save;
  public:
    static const RNG_type type=YARN2_t;
    const char * name(void);    
    void reset(void);
    void seed(long s=0l);
    long rand(void);    
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    YARN2 & YARN2::operator=(RNG<YARN2> &other); 
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a_1=1\,498\,809\,829\f$, \f$a_2=1\,160\,990\,996\f$ 
      and \f$m=2^{31}-1\f$ 
      \latexonly as proposed in \cite{lEcuyer1993}\endlatexonly. 
      This generator has a period of 
      \f$2^{2\cdot 31}-1\approx 2^{62}\approx 4.61\cdot 10^{18}\f$.
      The parameter \f$b\f$ is chosen to be \f$1\,235\,678\,93l\f$.
      \param a0_ multiplier \f$a_1\f$ 
      \param a1_ multiplier \f$a_2\f$ 
      \param modulus_ prime modulus
      \param seed_ default seed
      \param b_ parameter \f$b\f$
    */
    YARN2(long a0_=1498809829l, long a1_=1160990996l, 
	  long modulus_=2147483647l, 
	  long seed_=0l, long b_=123567893l);
  };

  inline long YARN2::rand(void) {
    long t=R.rand();
    return (t==0) ? 0 : b(t);
  }

  // -------------------------------------------------------------------

  //! YARN -- a modified multiple recursive generator
  /*!
    This multiple recursive generator uses a linear recurrence of order 
    three with a prime modulus.     
    \f[
    q_{i} = a_1\cdot q_{i-1}+a_2\cdot q_{i-2}+a_3\cdot q_{i-3}\mod m
    \f]
    Linear structures are destroyed by a nonlinear bijective mapping from
    \f$r_{i} = 
    \begin{cases}
    b^{q_{i}}\mod m & \text{if }q_{i}\ne 0 \\
    0         & \text{if }q_{i} =0 \\
    \end{cases}\,.
    \f$
    The parameter \f$b\f$ has to be a generating element of the 
    multiplicative group modulo \f$m\f$.
    \author Heiko Bauke
   */
  class YARN3 : public RNG<YARN3> {
  private:
    MRG3 R;
    power b;
    long b_save;
  public:
    static const RNG_type type=YARN3_t;
    const char * name(void);    
    void reset(void);
    void seed(long s=0l);
    long rand(void);    
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    YARN3 & YARN3::operator=(RNG<YARN3> &other); 
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a_1=2\,021\,422\,057\f$, \f$a_2=1\,826\,992\,351\f$,
      \f$a_3=1\,977\,753\,457\f$ and \f$m=2^{31}-1\f$ 
      \latexonly as proposed in \cite{lEcuyer1993}\endlatexonly. 
      This generator has a period of 
      \f$2^{3\cdot 31}-1\approx 2^{93}\approx 9.90\cdot 10^{27}\f$.
      The parameter \f$b\f$ is chosen to be \f$1\,235\,678\,93l\f$.
      \param a0_ multiplier \f$a_1\f$ 
      \param a1_ multiplier \f$a_2\f$ 
      \param a2_ multiplier \f$a_3\f$ 
      \param modulus_ prime modulus
      \param seed_ default seed
      \param b_ parameter \f$b\f$
    */
    YARN3(long a0_=2021422057l, long a1_=1826992351l, 
	  long a2_=1977753457l, long modulus_=2147483647l, 
	  long seed_=0l, long b_=123567893l);
  };

  inline long YARN3::rand(void) {
    long t=R.rand();
    return (t==0) ? 0 : b(t);
  }

  // -------------------------------------------------------------------

  //! YARN -- a modified multiple recursive generator
  /*!
    This multiple recursive generator uses a linear recurrence of order 
    four with a prime modulus.     
    \f[
    q_{i} = a_1\cdot q_{i-1}+a_2\cdot q_{i-2}+
    a_3\cdot q_{i-3}+a_4\cdot q_{i-4}\mod m
    \f]
    Linear structures are destroyed by a nonlinear bijective mapping from
    \f$r_{i} = 
    \begin{cases}
    b^{q_{i}}\mod m & \text{if }q_{i}\ne 0 \\
    0         & \text{if }q_{i} =0 \\
    \end{cases}\,.
    \f$
    The parameter \f$b\f$ has to be a generating element of the 
    multiplicative group modulo \f$m\f$.
    \author Heiko Bauke
   */
  class YARN4 : public RNG<YARN4> {
  private:
    MRG4 R;
    power b;
    long b_save;
  public:
    static const RNG_type type=YARN4_t;
    const char * name(void);    
    void reset(void);
    void seed(long s=0l);
    long rand(void);    
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    YARN4 & YARN4::operator=(RNG<YARN4> &other); 
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with \f$a_1=2\,001\,982\,722\f$, \f$a_2=1\,412\,284\,257\f$,
      \f$a_3=1\,155\,380\,217\f$, \f$a_4=1\,668\,339\,922\f$ and 
      \f$m=2^{31}-1\f$ 
      \latexonly as proposed by \cite{lEcuyer1993}\endlatexonly.
      This generator has a period of 
      \f$2^{4\cdot 31}-1\approx 2^{124}\approx 2.13\cdot 10^{37}\f$.
      The parameter \f$b\f$ is chosen to be \f$1\,235\,678\,93l\f$.
      \param a0_ multiplier \f$a_1\f$ 
      \param a1_ multiplier \f$a_2\f$ 
      \param a2_ multiplier \f$a_3\f$ 
      \param a3_ multiplier \f$a_4\f$ 
      \param modulus_ prime modulus
      \param seed_ default seed
      \param b_ parameter \f$b\f$
    */
    YARN4(long a0_=2001982722l, long a1_=1412284257l, 
	  long a2_=1155380217l, long a3_=1668339922l,
	  long modulus_=2147483647l, long seed_=0l, long b_=123567893l);
  };

  inline long YARN4::rand(void) {
    long t=R.rand();
    return (t==0) ? 0 : b(t);
  }

  // -------------------------------------------------------------------

  //! combined generator
  /*!
    This is a combined linear congruential random number generator 
    with two generators. 
    \f[
    \begin{split}
    q_{1,i} & =  a_1\cdot q_{1,i-1} \mod m_1 \\
    q_{2,i} & =  a_2\cdot q_{2,i-1} \mod m_2 \\
    r_i     & = q_{1,i}+q_{2,i} \mod m_1-1
    \end{split}
    \f]
    \latexonly See also \cite{lEcuyer1988}.\endlatexonly
    \author Heiko Bauke
  */
  class CLCG2 : public RNG<CLCG2> {
  private:
    generic_MLCG R1;
    generic_MLCG R2;
    long modulus;
  public:
    static const RNG_type type=CLCG2_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);
    long rand(void);    
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);    
    void load_status(const std::vector<long> &s);    
    CLCG2 & CLCG2::operator=(RNG<CLCG2> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with
      \f[
      \begin{split}
      q_{1,i} & =    376\,555\,083\cdot q_{1,i-1} \mod 2\,147\,482\,951 \\
      q_{2,i} & = 1\,028\,879\,659\cdot q_{2,i-1} \mod 2\,147\,482\,949 \\
      r_i     & = q_{1,i}+q_{2,i} \mod 2\,147\,482\,950\,.
      \end{split}
      \f]
      This generator has a period of 
      \f$\frac{(2\,147\,482\,951-1)(2\,147\,482\,949-1)}{2}\approx 2^{61} 
      \approx 2.31\cdot 10^{18}\f$. The multipliers were found by an exhausive 
      search applying the spectral test in up to eight dimensions.
      \param a1 multiplier \f$a_1\f$ 
      \param m1 prime modulus \f$m_1\f$ 
      \param a2 multiplier \f$a_2\f$ 
      \param m2 prime modulus \f$m_2\f$ 
      \param seed_ default seed
    */
    CLCG2(long a1= 376555083l, long m1=2147482951l,
	  long a2=1028879659l, long m2=2147482949l, 
	  long seed_=0l);    
  };

  inline long CLCG2::rand(void) {
    return static_cast<long>((static_cast<long long>(R1.rand())+
			      static_cast<long long>(R2.rand()))%
			     modulus);
  }
  
  // -------------------------------------------------------------------

  //! combined generator
  /*!
    This is a combined linear congruential random number generator 
    with three generators. 
    \f[
    \begin{split}
    q_{1,i} & =  a_1\cdot q_{1,i-1} \mod m_1 \\
    q_{2,i} & =  a_2\cdot q_{2,i-1} \mod m_2 \\
    q_{3,i} & =  a_3\cdot q_{3,i-1} \mod m_3 \\
    r_i     & = q_{1,i}+q_{2,i}+q_{3,i} \mod m_1-1
    \end{split}
    \f]
    \latexonly See also \cite{lEcuyer1988}.\endlatexonly
    \author Heiko Bauke
  */
  class CLCG3 : public RNG<CLCG3> {
  private:
    generic_MLCG R1;
    generic_MLCG R2;
    generic_MLCG R3;
    long modulus;
  public:
    static const RNG_type type=CLCG3_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);
    long rand(void);
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    CLCG3 & CLCG3::operator=(RNG<CLCG3> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with
      \f[
      \begin{split}
      q_{1,i} & =    376\,555\,083\cdot q_{1,i-1} \mod 2\,147\,482\,951 \\
      q_{2,i} & = 1\,028\,879\,659\cdot q_{2,i-1} \mod 2\,147\,482\,949 \\
      q_{3,i} & =    225\,802\,979\cdot q_{3,i-1} \mod 2\,147\,482\,943 \\
      r_i     & = q_{1,i}+q_{2,i}+q_{3,i} \mod 2\,147\,482\,950\,.
      \end{split}
      \f]
      This generator has a period of 
      \f$\frac{(2\,147\,482\,951-1)(2\,147\,482\,949-1)
      (2\,147\,482\,943-1)}{4}\approx 2^{91} \approx 2.48\cdot 10^{27}\f$.
      The multipliers were found by an exhausive 
      search applying the spectral test in up to eight dimensions.
      \param a1 multiplier \f$a_1\f$ 
      \param m1 prime modulus \f$m_1\f$ 
      \param a2 multiplier \f$a_2\f$ 
      \param m2 prime modulus \f$m_2\f$ 
      \param a3 multiplier \f$a_3\f$ 
      \param m3 prime modulus \f$m_3\f$ 
      \param seed_ default seed
    */
    CLCG3(long a1= 376555083l, long m1=2147482951l,
	  long a2=1028879659l, long m2=2147482949l,
	  long a3= 225802979l, long m3=2147482943l,
	  long seed_=0l);    
  };

  inline long CLCG3::rand(void) {
    return static_cast<long>((static_cast<long long>(R1.rand())+
			      static_cast<long long>(R2.rand())+
			      static_cast<long long>(R3.rand()))%
			   modulus);
  }
  
  // -------------------------------------------------------------------

  //! combined generator
  /*!
    This is a combined linear congruential random number generator 
    with four generators. 
    \f[
    \begin{split}
    q_{1,i} & =  a_1\cdot q_{1,i-1} \mod m_1 \\
    q_{2,i} & =  a_2\cdot q_{2,i-1} \mod m_2 \\
    q_{3,i} & =  a_3\cdot q_{3,i-1} \mod m_3 \\
    q_{4,i} & =  a_3\cdot q_{4,i-1} \mod m_4 \\
    r_i     & = q_{1,i}+q_{2,i}+q_{3,i}+q_{4,i} \mod m_1-1
    \end{split}
    \f]
    \latexonly See also \cite{lEcuyer1988}.\endlatexonly
    \author Heiko Bauke
  */
  class CLCG4 : public RNG<CLCG4> {
  private:
    generic_MLCG R1;
    generic_MLCG R2;
    generic_MLCG R3;
    generic_MLCG R4;
    long modulus;
  public:
    static const RNG_type type=CLCG4_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);
    long rand(void);
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);  
    void load_status(const std::vector<long> &s);
    CLCG4 & CLCG4::operator=(RNG<CLCG4> &other);
    //! constructor
    /*!
      The constructor's default values implement a pseudo random number 
      generator with
      \f[
      \begin{split}
      q_{1,i} & =    376\,555\,083\cdot q_{1,i-1} \mod 2\,147\,482\,951 \\
      q_{2,i} & = 1\,028\,879\,659\cdot q_{2,i-1} \mod 2\,147\,482\,949 \\
      q_{3,i} & =    225\,802\,979\cdot q_{3,i-1} \mod 2\,147\,482\,943 \\
      q_{4,i} & = 2\,028\,073\,966\cdot q_{3,i-1} \mod 2\,147\,482\,859 \\
      r_i     & = q_{1,i}+q_{2,i}+q_{3,i}+q_{4,i} \mod 2\,147\,482\,950\,.
      \end{split}
      \f]
      This generator has a period of 
      \f$\frac{(2\,147\,482\,951-1)(2\,147\,482\,949-1)
      (2\,147\,482\,943-1)(2\,147\,482\,859-1)}{8}\approx 2^{121}
      \approx 2.66\cdot 10^{36}\f$.
      The multipliers were found by an exhausive 
      search applying the spectral test in up to eight dimensions.
      \param a1 multiplier \f$a_1\f$ 
      \param m1 prime modulus \f$m_1\f$ 
      \param a2 multiplier \f$a_2\f$ 
      \param m2 prime modulus \f$m_2\f$ 
      \param a3 multiplier \f$a_3\f$ 
      \param m3 prime modulus \f$m_3\f$ 
      \param a4 multiplier \f$a_4\f$ 
      \param m4 prime modulus \f$m_4\f$ 
      \param seed_ default seed
    */
    CLCG4(long a1= 376555083l, long m1=2147482951l,
	  long a2=1028879659l, long m2=2147482949l,
	  long a3= 225802979l, long m3=2147482943l,
	  long a4=2028073966l, long m4=2147482859l,
	  long seed_=0l);    
  };

  inline long CLCG4::rand(void) {
    return static_cast<long>((static_cast<long long>(R1.rand())+
			      static_cast<long long>(R2.rand())+
			      static_cast<long long>(R3.rand())+
			      static_cast<long long>(R4.rand()))%
			     modulus);
  }
    
  // -------------------------------------------------------------------

  //! explicit inversive congruential generator
  /*!
    This is an explicit inversive congruential generator.
    \f[
    r_i=\overline{a\cdot i}\mod (2^{30}+2^{28}+3)
    \f]
    This generator has a period of 
    \f$2^{30}+2^{28}+3\approx 2^{30}\approx 1.34\cdot 10^{9}\f$.
    This type has excelent statistical properties but its period is
    too short for large applications. You may combine this genrator 
    with an other one.
    \latexonly See also \cite{Eichenauer1993}.\endlatexonly
    \author Heiko Bauke
  */
  class EINV : public RNG<EINV> {
  private:
    long a, i, di, modulus, a_save;
    long modulo_invers(const long a);
  public:
    static const RNG_type type=EINV_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);
    long rand(void);
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);
    EINV & EINV::operator=(RNG<EINV> &other);
    //! constructor
    /*!
      The default multiplier is \f$1\,073\,741\,831\f$.
      \param a_ multiplier \f$a\f$
      \param seed_ seed
    */
    EINV(long a_=1073741831l, long seed_=0l);    
  };

  inline long EINV::rand(void) {
    i=static_cast<long>
      ((static_cast<long long>(i)+
	static_cast<long long>(di))%modulus);
    long x=static_cast<long>((static_cast<long long>(a)*
			      static_cast<long long>(i))%modulus);
    // invert x
    long temp, q, flast=0, f=1, m=modulus;
    while (x>1) {
      temp=m%x; 
      q=m/x;
      m=x;  
      x=temp;  
      temp=f; 
      f=flast-q*f; 
      flast=temp;
    }
    return a==0 ? 0 : f>=0 ? f : f+modulus;
  }
  
  // -------------------------------------------------------------------

  //! combined generator
  /*!
    This is a combined generator.
    \f[
    \begin{split}
    q_{1,i}&=\overline{1073741831\cdot i}\mod (2^{30}+2^{28}+3) \\
    q_{2,i}&=\left\lfloor \frac{q_{3,i}}{2^{33}} \right\rfloor \\
    q_{3,i}&=18\,145\,460\,002\,477\,866\,997\cdot q_{3,i-1}+1\mod 2^{64} \\
    r_{i}&=q_{1,i}+q_{2,i} \mod 2^{31}
    \end{split}
    \f]
    \author Heiko Bauke
  */
  class EINVLCG64 : public RNG<EINVLCG64> {
  private:
    LCG64 R1;
    EINV R2;
  public:
    static const RNG_type type=EINVLCG64_t;
    const char * name(void);
    void reset(void);
    void seed(long s=0l);    
    long rand(void);
    void split(long s, long n);
    void jump2(long s);
    void save_status(std::vector<long> &s);
    void load_status(const std::vector<long> &s);    
    EINVLCG64 & EINVLCG64::operator=(RNG<EINVLCG64> &other);
    //! constructor
    /*!
      \param seed_ seed	
    */
    EINVLCG64(long seed_=0l);    
  };

  inline long EINVLCG64::rand(void) {
    return (R1.rand()+R2.rand())&0x7fffffffl;
  }

}


#endif
