// ---------------------------------------------------------------------
// Time-stamp: <Samstag, 25.01.2003, 14:00:07; edited by heiko>
// 
// Tina's random number generators TRNG
//
// useful functions for Tina's random number generators 
//
// Copyright (C) 2001, 2002 Heiko Bauke
//
// heiko.bauke@physik.uni-magdeburg.de
//
// TRNG is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation. This program
// is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// ---------------------------------------------------------------------

#if !defined TRNGLIB_H

#define TRNGLIB_H

#include <vector>
#include <stdexcept>
#include <string>

namespace TRNG {

  //! class for %error handling.
  /*!
    This class is thrown, if an %error occurs. 
  */
  class error : public std::exception {
    std::string _what;
  public:
    error(const std::string &what_arg) : _what(what_arg) {
    };
    const char * what(void) const {
      return _what.c_str();
    };
  };

  //! power
  /*!
    This data structure is used in the YARN random number generator family. 
    It allows the fast calculation of all powers of a given base \f$b\f$ 
    modulo \f$m\f$.
  */
  class power {
    long b, modulus;
    std::vector<long> b_power1, b_power0;
  
    void calc_b_power(void) {
      long long c=1ll;
      for (long i=0l; i<0x10000l; ++i) {
	b_power0[i]=c;
	c*=b;
	c%=modulus;
      }
      long long b_2_16=b;
      for (long i=0; i<16; ++i) {
	b_2_16*=b_2_16;
	b_2_16%=modulus;
      }
      c=1ll;
      for (long i=0l; i<0x10000l; ++i) {
	b_power1[i]=c;
	c*=b_2_16;
	c%=modulus;
      }
    }

  public:
    long operator()(const long n) const {
      return static_cast<long int>((static_cast<long long>(b_power1[n>>16])*
				    static_cast<long long>(b_power0[n&0xffff]))
				   %modulus);
    }
    
    long operator()(void) const {
      return b;
    }

    void set(const long b_, const long modulus_) {
      b=b_;
      modulus=modulus_;
      calc_b_power();
    }
    
    power(const long b_, const long modulus_) : 
      b_power0(0x10000), 
      b_power1(0x10000),
      b(b_),
      modulus(modulus_) {
      calc_b_power();
    }

    virtual ~power() {
    }
  };

  const char * version(void);
  long modulo_invers(long, long);
  void gauss(std::vector<long> &, std::vector<long> &, long);
  void matrix_mult(const std::vector<long> &, const std::vector<long> &,
		   std::vector<long> &, long);
  void matrix_vec_mult(const std::vector<long> &, const std::vector<long> &, 
		       std::vector<long> &, long);

  double Gamma(double);
  double ln_Gamma(double);
  double Gamma_P(double, double);
  double Gamma_Q(double, double);
  double incomp_Gamma(double, double);
  double comp_incomp_Gamma(double, double);
  double Gamma_ser(double, double);
  double Gamma_cf(double, double);
  double ln_factorial(long);
  long binomial_coeff(long, long);
  double errf(double);
  double chi_square_test(const std::vector<double> &, 
			 const std::vector<double> &);
  double chi_square_prob(double, long);
  double Stirling_num2(long, long);
  double Student_t(double, long, bool=true);
  long find_interval(const std::vector<double> &, const double);

  double uniform_pdf(double);
  double uniform_pdf(double, double, double);
  double uniformco_pdf(double);
  double uniformco_pdf(double, double, double);
  double uniformcc_pdf(double);
  double uniformcc_pdf(double, double, double);
  double uniformoc_pdf(double);
  double uniformoc_pdf(double, double, double);
  double uniformoo_pdf(double);
  double uniformoo_pdf(double, double, double);
  double normal_dist_pdf(double, double, double);
  double exp_dist_pdf(double, double);
  double laplace_dist_pdf(double, double);
  double tent_dist_pdf(double, double);
  double Gamma_dist_pdf(double, double, double);
  double Beta_dist_pdf(double, double, double);
  double chi_square_dist_pdf(double, double);
  double Student_t_dist_pdf(double, double);
  double binomial_dist_pdf(long, long, double);
  double poisson_dist_pdf(long, double);
  double geometric_dist_pdf(long, double);

}
  
#endif


